unit AlphaEffects;

interface

{$WARNINGS OFF}

uses
  //Classes, SysUtils, Windows, Forms, Messages, Graphics, ExtCtrls, Math;
  Kol, windows, VCLType, Share, Debug;

type
  TAlphaTable = array[0..255] of Byte;

  TAlphaFXBase = class(TDebug)
  private
    FRunning: Boolean;
    FRunTimeMS: Integer;
    FForm: PControl;
    FWait: Boolean;
    FEnd: Boolean;
    FRestore: Boolean;

    FOrgBMP: TBitmap;

    FOnStart: TNotifyEvent;
    FOnEnd: TNotifyEvent;

    Blend: TBlendFunction;
    BmpTopLeft,
    TopLeft: TPoint;
    BmpSize: TSize;

    FTimer: TTimer;

    StartTicks: Cardinal;

    procedure DoStep;
    procedure CopyBMP;

    procedure OnTimer(Sender: PObj);
  protected
    FReverse: Boolean;
    FStartValue1, FEndValue1: Integer;
    FStartValue2, FEndValue2: Integer;
    FStartValue3, FEndValue3: Integer;
    FStartValue4, FEndValue4: Integer;
    FWorkBMP: TBitmap;

    procedure Start; virtual; abstract;
    procedure Step(Value1: Integer; Value2: Integer; Value3, Value4: Integer); virtual; abstract;

    procedure Done; virtual;
  public
    constructor Create(Form: PControl; RunTimeMS: Integer; Restore: Boolean = True; Reverse: Boolean = False; Wait: Boolean = True; OnStart: TNotifyEvent = nil; OnEnd: TNotifyEvent = nil);
    destructor Destroy; override;
  end;

  TAlphaFXEffect = (afxLeftRight, afxRightLeft, afxTopBottom, afxBottomTop,
                      afxDoubleInwardsVertical, afxDoubleOutwardsVertical,
                      afxDoubleInwardsHorizontal, afxDoubleOutwardsHorizontal,
                      afxFade,
                      afxRadarRandom,
                      afxCustom,
                      afxRadarCounterClockwise0, afxRadarCounterClockwise90,
                      afxRadarCounterClockwise180, afxRadarCounterClockwise270,
                      afxRadarClockwise0, afxRadarClockwise90,
                      afxRadarClockwise180, afxRadarClockwise270
                    );

  TAlphaFXCustomEffect = Array of TAlphaFXEffect;

  TAlphaFX = class(TAlphaFXBase)
  private
    FProgress, WrapValue1, WrapValue2: Integer;
    FEffect: TAlphaFXEffect;
    FCount: Integer;
    FCustomValues: TAlphaFXCustomEffect;

    FCircleTable: array of Array of Word;
    FStartAngle: Word;
    FClockwise: Boolean;

    procedure SetAlpha(StartX, StartY, EndX, EndY: Integer; Alpha: Byte);

    procedure CreateCircleTable(Width, Height: Integer);
    function GetAngleTable(HalfWidth, HalfHeight, X, Y: Integer): Word;
    procedure DoRadar(Value: Integer);
  protected
    procedure Start; override;
    procedure Step(Value1, Value2, Value3, Value4: Integer); override;
  public
    constructor Create(Form: PControl; Effect: TAlphaFXEffect; Count: Integer; RunTimeMS: Integer; Custom: TAlphaFXCustomEffect = nil; Restore: Boolean = True; Reverse: Boolean = False; Wait: Boolean = True; OnStart: TNotifyEvent = nil; OnEnd: TNotifyEvent = nil);
  end;

  var PreAlphaTable: Array[0..255] of TAlphaTable;

  procedure AlphaFX(Form: PControl; Effect: TAlphaFXCustomEffect; Count: Integer; RunTimeMS: Integer; Restore: Boolean = True; Reverse: Boolean = False; OnStart: TNotifyEvent = nil; OnEnd: TNotifyEvent = nil);
  function  afx(CustomValues: Array of TAlphaFXEffect): TAlphaFXCustomEffect;
  function  afx_Random(Length: Cardinal): TAlphaFXCustomEffect;
  function  afx_LeftRight: TAlphaFXCustomEffect;
  function  afx_RightLeft: TAlphaFXCustomEffect;
  function  afx_TopBottom: TAlphaFXCustomEffect;
  function  afx_BottomTop: TAlphaFXCustomEffect;
  function  afx_Clockwise: TAlphaFXCustomEffect;
  function  afx_CounterClockwise: TAlphaFXCustomEffect;
  function  afx_Outwards: TAlphaFXCustomEffect;
  function  afx_Inwards: TAlphaFXCustomEffect;
  function  afx_DoubleHorizontal: TAlphaFXCustomEffect;
  function  afx_DoubleVertical: TAlphaFXCustomEffect;
  function  afx_RadarSweep0: TAlphaFXCustomEffect;
  function  afx_RadarSweep90: TAlphaFXCustomEffect;
  function  afx_RadarSweep180: TAlphaFXCustomEffect;
  function  afx_RadarSweep270: TAlphaFXCustomEffect;

  function UpdateLayeredWindow( hwnd: HWND; hdcDst: HDC; pptDst: PPoint; psize: PSize;
                                hdcSrc: HDC; pptSrc: PPoint; crKey: TColor;
                                blend: PBlendFunction; dwFlags: Dword): Boolean;
                                stdcall; external 'User32.dll' name 'UpdateLayeredWindow';

  function PrintWindow(HWND:HWND;hdcBlt:HDC;nFlags:DWORD):BOOL; stdcall; external 'user32.dll';

implementation

const
  WS_EX_LAYERED = $00080000;
  AC_SRC_OVER   = $00;
  AC_SRC_ALPHA  = $01;
  ULW_ALPHA     = $00000002;
  ULW_OPAQUE    = $00000004;


{ Helper procs }

function Bounds(ALeft, ATop, AWidth, AHeight: Integer): TRect;
begin
  with Result do
  begin
    Left := ALeft;
    Top := ATop;
    Right := ALeft + AWidth;
    Bottom :=  ATop + AHeight;
  end;
end;

function GetRadarRandom: TAlphaFXEffect;
var i: Integer;
begin
  Result := afxRadarCounterClockwise0;

  Randomize;
  i := Random(8);

  case i of
    0: Result := afxRadarCounterClockwise0;
    1: Result := afxRadarCounterClockwise90;
    2: Result := afxRadarCounterClockwise180;
    3: Result := afxRadarCounterClockwise270;
    4: Result := afxRadarClockwise0;
    5: Result := afxRadarClockwise90;
    6: Result := afxRadarClockwise180;
    7: Result := afxRadarClockwise270;
  end;
end;

procedure AlphaFX(Form: PControl; Effect: TAlphaFXCustomEffect; Count: Integer; RunTimeMS: Integer; Restore: Boolean = True; Reverse: Boolean = False; OnStart: TNotifyEvent = nil; OnEnd: TNotifyEvent = nil);
begin
  with TAlphaFX.Create(Form, afxCustom, Count, RunTimeMS, Effect, Restore, Reverse, True, OnStart, OnEnd) do Free;
end;

function afx(CustomValues: array of TAlphaFXEffect): TAlphaFXCustomEffect;
var i: Integer;
begin
  Result := nil;

  if Length(CustomValues) > 0 then begin
    SetLength(Result, Length(CustomValues));

    for i := 0 to Length(CustomValues) - 1 do
      Result[i] := CustomValues[i];
  end;
end;

function afx_Random(Length: Cardinal): TAlphaFXCustomEffect;
var I: Integer;
begin
  Result := nil;

  if Length > 0 then begin
    SetLength(Result, Length);

    for i := 0 to Length - 1 do begin
      Randomize;
      Result[i] := TAlphaFXEffect(Random(Ord(afxCustom)));
    end;
  end;
end;

function afx_LeftRight: TAlphaFXCustomEffect;
begin
  Result := afx([afxLeftRight, afxRightLeft]);
end;

function afx_RightLeft: TAlphaFXCustomEffect;
begin
  Result := afx([afxRightLeft, afxLeftRight]);
end;

function afx_TopBottom: TAlphaFXCustomEffect;
begin
  Result := afx([afxTopBottom, afxBottomTop]);
end;

function afx_BottomTop: TAlphaFXCustomEffect;
begin
  Result := afx([afxBottomTop, afxTopBottom]);
end;

function afx_Clockwise: TAlphaFXCustomEffect;
begin
  Result := afx([afxRightLeft, afxTopBottom, afxRightLeft, afxBottomTop]);
end;

function afx_CounterClockwise: TAlphaFXCustomEffect;
begin
  Result := afx([afxRightLeft, afxBottomTop, afxRightLeft, afxTopBottom]);
end;

function afx_Outwards: TAlphaFXCustomEffect;
begin
  Result := afx([afxDoubleOutwardsHorizontal, afxDoubleOutwardsVertical]);
end;

function afx_Inwards: TAlphaFXCustomEffect;
begin
  Result := afx([afxDoubleInwardsHorizontal, afxDoubleInwardsVertical]);
end;

function afx_DoubleHorizontal: TAlphaFXCustomEffect;
begin
  Result := afx([afxDoubleInwardsHorizontal, afxDoubleOutwardsHorizontal]);
end;

function afx_DoubleVertical: TAlphaFXCustomEffect;
begin
  Result := afx([afxDoubleInwardsVertical, afxDoubleOutwardsVertical]);
end;

function afx_RadarSweep0: TAlphaFXCustomEffect;
begin
  Result := afx([afxRadarCounterClockwise0, afxRadarClockwise0]);
end;

function afx_RadarSweep90: TAlphaFXCustomEffect;
begin
  Result := afx([afxRadarCounterClockwise90, afxRadarClockwise90]);
end;

function afx_RadarSweep180: TAlphaFXCustomEffect;
begin
  Result := afx([afxRadarCounterClockwise180, afxRadarClockwise180]);
end;

function afx_RadarSweep270: TAlphaFXCustomEffect;
begin
  Result := afx([afxRadarCounterClockwise270, afxRadarClockwise270]);
end;

procedure InitAlphaTables;
var i, j: Integer;
begin
  for i := 0 to 255 do
    for j := 0 to 255 do
      PreAlphaTable[i, j] := (j * i) shr 8;
end;

{ TAlphaFX }

procedure TAlphaFXBase.CopyBMP;
begin
  BitBlt(FWorkBMP.Canvas.Handle, 0, 0, FWorkBMP.Width, FWorkBMP.Height, FOrgBMP.Canvas.Handle, 0, 0, SRCCOPY);
end;

constructor TAlphaFXBase.Create(Form: PControl; RunTimeMS: Integer; Restore: Boolean; Reverse: Boolean; Wait: Boolean; OnStart, OnEnd: TNotifyEvent);
var
  DC: HDC;
  WS: Cardinal;
  R: TRect;
begin
  inherited Create;//(Form);
  FRestore := Restore;
  FForm := Form;

  FRunTimeMS := RunTimeMS;
  FReverse := Reverse;
  FRunning := False;
  FEnd := False;
  FWait := Wait;

  GetWindowRect(FForm.Handle, R);
  FOrgBMP := NewBitmap(R.Right - R.Left,R.Bottom - R.Top);//TBitmap.Create;
  FOrgBMP.PixelFormat := pf32Bit;
  FWorkBMP := NewBitmap(R.Right - R.Left,R.Bottom - R.Top);//TBitmap.Create;
  FWorkBMP.PixelFormat := pf32Bit;//FOrgBMP.PixelFormat;

  FOnStart := OnStart;
  FOnEnd := OnEnd;

  WS := GetWindowLong(FForm.Handle, GWL_EXSTYLE);
  SetWindowLong(FForm.Handle, GWL_EXSTYLE, WS or WS_EX_LAYERED);

  BmpSize.cx := R.Right - R.Left;
  BmpSize.cy := R.Bottom - R.Top;

  //FOrgBMP.Canvas.CopyRect(Bounds(0, 0, BmpSize.cx, BmpSize.cy), FForm.{GetFormImage.}Canvas, Bounds(0, 0, BmpSize.cx, BmpSize.cy));
  PrintWindow(FForm.Handle, FOrgBMP.Canvas.Handle, 0);

  GetWindowRect(FForm.Handle, R);
  TopLeft.X := R.Left;
  TopLeft.Y := R.Top;
  BmpTopLeft := Point(0, 0);

  FRunning := True;

  FStartValue1 := 0;
  FEndValue1 := 0;
  FStartValue2 := 0;
  FEndValue2 := 0;
  FStartValue3 := 0;
  FEndValue3 := 0;

  CopyBMP;
  Start;

  if Assigned(FOnStart) then
    FOnStart({Self}FForm);

  StartTicks := GetTickCount;

  DC := GetDC(0);

  with Blend do
  begin
    BlendOp :=             AC_SRC_OVER;
    BlendFlags :=          0;
    SourceConstantAlpha := 255;
    AlphaFormat :=         AC_SRC_ALPHA;
  end;

  UpdateLayeredWindow(FForm.Handle, DC, @TopLeft, @BmpSize, FWorkBmp.Canvas.Handle, @BmpTopLeft, clNone, @Blend, ULW_ALPHA);

  ReleaseDC(0, DC);

  DoStep;

  if FWait then begin
    repeat
      //Applet.HandleMessage; //!!!!!!!!!!
      if not Applet.ProcessMessage then WaitMessage;
    until FEnd = True;
    Done;
  end;
end;

destructor TAlphaFXBase.Destroy;
begin
  FOrgBMP.Free;
  FWorkBMP.Free;

  inherited;
end;

procedure TAlphaFXBase.Done;
var WS: Cardinal;
    DC: hDC;
begin
  if Assigned(FOnEnd) then
    FOnEnd({Self}FForm);

  if FRestore then begin
    CopyBMP;

    DC := GetDC(0);
    UpdateLayeredWindow(FForm.Handle, DC, @TopLeft, @BmpSize, FWorkBmp.Canvas.Handle, @BmpTopLeft, clNone, @Blend, ULW_OPAQUE);
    ReleaseDC(0, DC);

    WS := GetWindowLong(FForm.Handle, GWL_EXSTYLE);
    if WS and WS_EX_LAYERED = WS_EX_LAYERED then
      SetWindowLong(FForm.Handle, GWL_EXSTYLE, WS - WS_EX_LAYERED);
  end;
end;

procedure TAlphaFXBase.DoStep;
var DC: hDC;
    Running: Integer;

    Value1, Value2, Value3, Value4: Integer;
begin
  Running := GetTickCount - StartTicks;

  FEnd := (Running > FRunTimeMS);

  if FEnd then begin
    if not FWait then Done;
  end else begin
    CopyBMP;

    if FStartValue1 < FEndValue1 then begin
      Value1 := Round(FStartValue1 + ((FEndValue1 - FStartValue1) * (Running / FRunTimeMS)));
    end else begin
      Value1 := Round(FStartValue1 - ((FStartValue1 - FEndValue1) * (Running / FRunTimeMS)));
    end;

    if FStartValue2 < FEndValue2 then begin
      Value2 := Round(FStartValue2 + ((FEndValue2 - FStartValue2) * (Running / FRunTimeMS)));
    end else begin
      Value2 := Round(FStartValue2 - ((FStartValue2 - FEndValue2) * (Running / FRunTimeMS)));
    end;

    if FStartValue3 < FEndValue3 then begin
      Value3 := Round(FStartValue3 + ((FEndValue3 - FStartValue3) * (Running / FRunTimeMS)));
    end else begin
      Value3 := Round(FStartValue3 - ((FStartValue3 - FEndValue3) * (Running / FRunTimeMS)));
    end;

    if FStartValue4 < FEndValue4 then begin
      Value4 := Round(FStartValue4 + ((FEndValue4 - FStartValue4) * (Running / FRunTimeMS)));
    end else begin
      Value4 := Round(FStartValue4 - ((FStartValue4 - FEndValue4) * (Running / FRunTimeMS)));
    end;

    Step(Value1, Value2, Value3, Value4);

    DC := GetDC(0);
    UpdateLayeredWindow(FForm.Handle, DC, @TopLeft, @BmpSize, FWorkBmp.Canvas.Handle, @BmpTopLeft, clNone, @Blend, ULW_ALPHA);
    ReleaseDC(0, DC);

    FTimer := NewTimer(1);//TTimer.Create(Self);
    FTimer.Interval := 1;
    FTimer.OnTimer := OnTimer;
    FTimer.Enabled := True;
  end;
end;

procedure TAlphaFXBase.OnTimer(Sender: PObj);
begin
  FTimer.Free;
  DoStep;
end;

{ TAlphaFX }

constructor TAlphaFX.Create(Form: PControl; Effect: TAlphaFXEffect;
  Count: Integer; RunTimeMS: Integer;  Custom: TAlphaFXCustomEffect;
  Restore: Boolean; Reverse: Boolean; Wait: Boolean; OnStart: TNotifyEvent;
  OnEnd: TNotifyEvent);
begin
  FEffect := Effect;
  FCount := Count;

  if Effect = afxCustom then begin
    if Custom <> nil then
      FCustomValues := Custom
    else
      FCustomValues := afx_Random(FCount);
  end;

  inherited Create(Form, RunTimeMS, Restore, Reverse, Wait, OnStart, OnEnd);
end;

procedure TAlphaFX.DoRadar(Value: Integer);
type TRGBAData = Array[0..3] of Byte;
     PRGBAData = ^TRGBAData;
var i, j: Integer;
    Pixel: PRGBAData;
    Base: Pointer;
    BaseInc: Cardinal;
    AlphaTable1, AlphaTable2: TAlphaTable;
    HalfWidth, HalfHeight: Integer;

    Angle, CurAngle, SaveAngle: Integer;
    CurAlpha, Alpha1, Alpha2: Integer;
begin
  HalfWidth := FWorkBMP.Width div 2;
  HalfHeight := FWorkBMP.Height div 2;

  Angle := (Value mod 360);

  i := Value div 360;
  if not FReverse then begin
    Alpha1 := Round((255 / (FCount)) * i);
    Alpha2 := Round((255 / (FCount)) * (i + 1));
  end else begin
    Alpha1 := 255 - Round((255 / (FCount)) * i);
    Alpha2 := 255 - Round((255 / (FCount)) * (i + 1));
  end;

  AlphaTable1 := PreAlphaTable[Alpha1];
  AlphaTable2 := PreAlphaTable[Alpha2];

  Base := FWorkBMP.ScanLine[0];
  BaseInc := Cardinal(FWorkBMP.ScanLine[1]) - Cardinal(FWorkBMP.ScanLine[0]);

  for i := 0 to FWorkBMP.Height - 1 do begin
    Cardinal(Pixel) := Cardinal(Base);

    for j := 0 to FWorkBMP.Width - 1 do begin
      if FClockwise then
        CurAngle := 360 - GetAngleTable(HalfWidth, HalfHeight, j, i)
      else
        CurAngle := GetAngleTable(HalfWidth, HalfHeight, j, i);

      SaveAngle := Angle;

      if ((CurAngle < 4) AND (Angle <= 8)) OR
         ((CurAngle >= 356) AND (Angle >= 352)) then begin
        Angle := Angle + 16;
        CurAngle := CurAngle + 16;
      end;

      if (CurAngle < 4) AND (Angle > 8) then begin
        CurAlpha := Alpha1 + ((Alpha2 - Alpha1) shr 3) * (CurAngle + 4);

        Pixel[0] := PreAlphaTable[CurAlpha, Pixel[0]];
        Pixel[1] := PreAlphaTable[CurAlpha, Pixel[1]];
        Pixel[2] := PreAlphaTable[CurAlpha, Pixel[2]];
        Pixel[3] := CurAlpha;
      end else if (CurAngle >= 356) AND (Angle <= 352) then begin
        CurAlpha := Alpha1 + ((Alpha2 - Alpha1) shr 3) * (CurAngle - 356);

        Pixel[0] := PreAlphaTable[CurAlpha, Pixel[0]];
        Pixel[1] := PreAlphaTable[CurAlpha, Pixel[1]];
        Pixel[2] := PreAlphaTable[CurAlpha, Pixel[2]];
        Pixel[3] := CurAlpha;
      end else if CurAngle <= Angle - 8 then begin
        Pixel[0] := AlphaTable2[Pixel[0]];
        Pixel[1] := AlphaTable2[Pixel[1]];
        Pixel[2] := AlphaTable2[Pixel[2]];
        Pixel[3] := Alpha2;
      end else if CurAngle > Angle + 8 then begin
        Pixel[0] := AlphaTable1[Pixel[0]];
        Pixel[1] := AlphaTable1[Pixel[1]];
        Pixel[2] := AlphaTable1[Pixel[2]];
        Pixel[3] := Alpha1;
      end else begin
        CurAlpha := CurAngle - Angle;
        if CurAlpha >= 0 then begin
          CurAlpha := Alpha1 + (((255 - Alpha1) shr 3) * (8 - CurAlpha));

          Pixel[0] := PreAlphaTable[CurAlpha, Pixel[0]];
          Pixel[1] := PreAlphaTable[CurAlpha, Pixel[1]];
          Pixel[2] := PreAlphaTable[CurAlpha, Pixel[2]];
          Pixel[3] := CurAlpha;
        end else if CurAlpha < 0 then begin
          CurAlpha := Alpha2 + (((255 - Alpha2) shr 3) * (8 + CurAlpha));

          Pixel[0] := PreAlphaTable[CurAlpha, Pixel[0]];
          Pixel[1] := PreAlphaTable[CurAlpha, Pixel[1]];
          Pixel[2] := PreAlphaTable[CurAlpha, Pixel[2]];
          Pixel[3] := CurAlpha;
        end;
      end;

      Angle := SaveAngle;

      Cardinal(Pixel) := Cardinal(Pixel) + 4;
    end;

    Cardinal(Base) := Cardinal(Base) + BaseInc;
  end;
end;

procedure TAlphaFX.SetAlpha(StartX, StartY, EndX, EndY: Integer; Alpha: Byte);
type TRGBAData = Array[0..3] of Byte;
     PRGBAData = ^TRGBAData;
var i, j: Integer;
    Pixel: PRGBAData;
    Base: Pointer;
    BaseInc: Integer;
    AlphaTable: TAlphaTable;
begin
  if StartX < 0 then StartX := 0;
  if StartY < 0 then StartY := 0;
  if EndX > FWorkBMP.Width - 1 then EndX := FWorkBMP.Width - 1;
  if EndY > FWorkBMP.Height - 1 then EndY := FWorkBMP.Height - 1;

  if EndX - StartX < 0 then Exit;
  if EndY - StartY < 0 then Exit;

  AlphaTable := PreAlphaTable[Alpha];

  Base := FWorkBMP.ScanLine[StartY];
  if StartY < EndY then
    BaseInc := Cardinal(FWorkBMP.ScanLine[StartY + 1]) - Cardinal(FWorkBMP.ScanLine[StartY])
  else
    BaseInc := 0;

  for i := StartY to EndY do begin
    Cardinal(Pixel) := Cardinal(Base) + (StartX shl 2);

    for j := StartX to EndX do begin
      Pixel[0] := AlphaTable[Pixel[0]];
      Pixel[1] := AlphaTable[Pixel[1]];
      Pixel[2] := AlphaTable[Pixel[2]];
      Pixel[3] := Alpha;

      Cardinal(Pixel) := Cardinal(Pixel) + 4;
    end;

    Cardinal(Base) := Cardinal(Base) + BaseInc;
  end;
end;

procedure TAlphaFX.CreateCircleTable(Width, Height: Integer);
var M: Integer;
    i, j: Integer;
begin
  M := Max((Width + 2) div 2, (Height + 2) div 2);
  SetLength(FCircleTable, M, M);

  for i := 1 to M - 1 do
    for j := 1 to M - 1 do
      FCircleTable[i, j] := Round((ArcTan(j / i) / (2 * pi)) * 360);
end;

function TAlphaFX.GetAngleTable(HalfWidth, HalfHeight, X, Y: Integer): Word;
var Quadrant: Integer;
    A, B: Integer;
begin
  Quadrant := 0;
  Result := 0;

  if X = HalfWidth then begin
    if Y < HalfHeight then
      Quadrant := 1
    else
      Quadrant := 3;
  end else if X > HalfWidth then begin
    if Y <= HalfHeight then
      Quadrant := 0
    else
      Quadrant := 3;
  end else if X < HalfWidth then begin
    if Y < HalfHeight then
      Quadrant := 1
    else
      Quadrant := 2;
  end;

  if Quadrant = 0 then begin
    A := X - HalfWidth;
    B := HalfHeight - Y;

    Result := FCircleTable[A, B];
  end else if Quadrant = 1 then begin
    A := HalfHeight - Y;
    B := HalfWidth - X;

    Result := FCircleTable[A, B] + 90;
  end else if Quadrant = 2 then begin
    A := HalfWidth - X;
    B := Y - HalfHeight;

    Result := FCircleTable[A, B] + 180;
  end else if Quadrant = 3 then begin
    A := Y - HalfHeight;
    B := X - HalfWidth;

    Result := FCircleTable[A, B] + 270;
  end;

  Result := Result + 360 - FStartAngle;
  if Result >= 360 then Result := Result - 360;
end;

procedure TAlphaFX.Start;
var i: Integer;
begin
  //inherited; !!!!!!!!!!!!!!!!!!!!

  CreateCircleTable(FWorkBMP.Width, FWorkBMP.Height);

  if FEffect = afxRadarRandom then begin
    SetLength(FCustomValues, FCount);

    for i := 0 to FCount - 1 do
      FCustomValues[i] := GetRadarRandom;

    FEffect := afxCustom;
  end;

  if FEffect = afxCustom then begin
    if LEngth(FCustomValues) > 0 then

    for i := 0 to FCount - 1 do
      if FCustomValues[i] = afxRadarRandom then
        FCustomValues[i] := GetRadarRandom;
  end;

  WrapValue1 := FWorkBMP.Width + 16;
  WrapValue2 := FWorkBMP.Height + 16;

  FStartAngle := 0;

  FStartValue1 := 0;
  FEndValue1 := WrapValue1 * FCount;
  FStartValue2 := 0;
  FEndValue2 := WrapValue2 * FCount;
  FStartValue3 := 0;
  FEndValue3 := (360 * FCount) - 1;
  FStartValue4 := 0;
  FEndValue4 := 255;

  Step(FStartValue1, FStartValue2, FStartValue3, FStartValue4);
end;

procedure TAlphaFX.Step(Value1, Value2, Value3, Value4: Integer);
var Alpha1, Alpha2: Integer;
    i, j: Integer;
    Effect: TAlphaFXEffect;

    WrapValue, Value: Integer;
begin
  //inherited;  !!!!!!!!!!!!!!!!!!

  WrapValue := 0;
  Value := 0;

  Effect := FEffect;

  if (FEffect = afxCustom) then
    Effect := FCustomValues[(Value1 div WrapValue1) mod Length(FCustomValues)];

  { Radar }

  if Effect in [afxRadarCounterClockwise0, afxRadarCounterClockwise90, afxRadarCounterClockwise180, afxRadarCounterClockwise270,
                   afxRadarClockwise0, afxRadarClockwise90, afxRadarClockwise180, afxRadarClockwise270] then begin
    if Effect = afxRadarCounterClockwise0 then
      FStartAngle := 0;
    if Effect = afxRadarCounterClockwise90 then
      FStartAngle := 90;
    if Effect = afxRadarCounterClockwise180 then
      FStartAngle := 180;
    if Effect = afxRadarCounterClockwise270 then
      FStartAngle := 270;

    if Effect = afxRadarClockwise0 then
      FStartAngle := 0;
    if Effect = afxRadarClockwise90 then
      FStartAngle := 90;
    if Effect = afxRadarClockwise180 then
      FStartAngle := 180;
    if Effect = afxRadarClockwise270 then
      FStartAngle := 270;

    FClockwise := (Effect in [afxRadarClockwise0, afxRadarClockwise90, afxRadarClockwise180, afxRadarClockwise270]);

    DoRadar(Value3);

    Exit;
  end;

  { Fade }

  if Effect = afxFade then begin
    if not FReverse then
      Alpha1 := Value4
    else
      Alpha1 := 255 - Value4;

    SetAlpha(0, 0, FWorkBMP.Width - 1, FWorkBMP.Height - 1, Alpha1);

    Exit;
  end;

  { Sweep }

  if Effect in [afxLeftRight, afxRightLeft, afxDoubleInwardsHorizontal, afxDoubleOutwardsHorizontal] then begin
    Value := Value1;
    WrapValue := WrapValue1;
  end;

  if Effect in [afxTopBottom, afxBottomTop, afxDoubleInwardsVertical, afxDoubleOutwardsVertical] then begin
    Value := Value2;
    WrapValue := WrapValue2;
  end;

  FProgress := (Value mod WrapValue) - 8;
  if FCount > 1 then begin
    Alpha1 := Round(((Value div WrapValue) + 1) * (256 / FCount));
    if Alpha1 > 255 then Alpha1 := 255;

    If not FReverse then begin
      Alpha2 := Alpha1 - Round(256 / FCount);
    end else begin
      Alpha1 := 255 - Alpha1;
      Alpha2 := Alpha1 + Round(256 / FCount);
    end;
  end else begin
    if not FReverse then begin
      Alpha1 := 255;
      Alpha2 := 0;
    end else begin
      Alpha1 := 0;
      Alpha2 := 255;
    end;
  end;

  if Effect = afxLeftRight then begin
    SetAlpha(0, 0, FProgress - 9, FWorkBMP.Height - 1, Alpha1);
    for i := 0 to 7 do
      SetAlpha(FProgress + i - 8, 0, FProgress + i - 8, FWorkBMP.Height - 1, Alpha1 + Round(((255 - Alpha1) / 8) * (i + 1)));
    for i := 0 to 7 do
      SetAlpha(FProgress + i, 0, FProgress + i, FWorkBMP.Height - 1, Alpha2 + Round(((255 - Alpha2) / 8) * (8 - (i + 1))));
    SetAlpha(FProgress + 8, 0, FWorkBMP.Width, FWorkBMP.Height - 1, Alpha2);
  end;

  if Effect = afxRightLeft then begin
    FProgress := FWorkBMP.Width - 1 - FProgress;

    SetAlpha(0, 0, FProgress - 9, FWorkBMP.Height - 1, Alpha2);
    for i := 0 to 7 do
      SetAlpha(FProgress + i - 8, 0, FProgress + i - 8, FWorkBMP.Height - 1, Alpha2 + Round(((255 - Alpha2) / 8) * (i + 1)));
    for i := 0 to 7 do
      SetAlpha(FProgress + i, 0, FProgress + i, FWorkBMP.Height - 1, Alpha1 + Round(((255 - Alpha1) / 8) * (8 - (i + 1))));
    SetAlpha(FProgress + 8, 0, FWorkBMP.Width, FWorkBMP.Height - 1, Alpha1);
  end;

  if Effect = afxTopBottom then begin
    SetAlpha(0, 0, FWorkBMP.Width - 1, FProgress - 9, Alpha1);
    for i := 0 to 7 do
      SetAlpha(0, FProgress + i - 8, FWorkBMP.Width - 1, FProgress + i - 8, Alpha1 + Round(((255 - Alpha1) / 8) * (i + 1)));
    for i := 0 to 7 do
      SetAlpha(0, FProgress + i, FWorkBMP.Width - 1, FProgress + i, Alpha2 + Round(((255 - Alpha2) / 8) * (8 - (i + 1))));
    SetAlpha(0, FProgress + 8, FWorkBMP.Width - 1, FWorkBMP.Height - 1, Alpha2);
  end;

  if Effect = afxBottomTop then begin
    FProgress := FWorkBMP.Height - 1 - FProgress;

    SetAlpha(0, 0, FWorkBMP.Width - 1, FProgress - 9, Alpha2);
    for i := 0 to 7 do
      SetAlpha(0, FProgress + i - 8, FWorkBMP.Width - 1, FProgress + i - 8, Alpha2 + Round(((255 - Alpha2) / 8) * (i + 1)));
    for i := 0 to 7 do
      SetAlpha(0, FProgress + i, FWorkBMP.Width - 1, FProgress + i, Alpha1 + Round(((255 - Alpha1) / 8) * (8 - (i + 1))));
    SetAlpha(0, FProgress + 8, FWorkBMP.Width - 1, FWorkBMP.Height - 1, Alpha1);
  end;

  if Effect = afxDoubleInwardsVertical then begin
    FProgress := FProgress div 2;
    j := FWorkBMP.Height div 2;

    SetAlpha(0, 0, FWorkBMP.Width - 1, FProgress - 1 - 8, Alpha1);
    for i := 0 to 7 do
      SetAlpha(0, FProgress - 8 + i, FWorkBMP.Width - 1, FProgress - 8 + i, Alpha1 + Round(((255 - Alpha1) / 8) * (i + 1)));
    for i := 0 to 7 do
      if FProgress + i < j then
        SetAlpha(0, FProgress + i, FWorkBMP.Width - 1, FProgress + i, Alpha2 + Round(((255 - Alpha2) / 8) * (8 - (i + 1))));
    SetAlpha(0, FProgress + 8, FWorkBMP.Width - 1, j - 1, Alpha2);

    SetAlpha(0, j, FWorkBMP.Width - 1, FWorkBMP.Height - FProgress - 1 - 8, Alpha2);
    for i := 0 to 7 do
      if FWorkBMP.Height - FProgress - 8 + i >= j then
        SetAlpha(0, FWorkBMP.Height - FProgress - 8 + i, FWorkBMP.Width - 1, FWorkBMP.Height - FProgress - 8 + i, Alpha2 + Round(((255 - Alpha2) / 8) * (i + 1)));
    for i := 0 to 7 do
      SetAlpha(0, FWorkBMP.Height - FProgress + i, FWorkBMP.Width - 1, FWorkBMP.Height - FProgress + i, Alpha1 + Round(((255 - Alpha1) / 8) * (8 - (i + 1))));
    SetAlpha(0, FWorkBMP.Height - FProgress + 8, FWorkBMP.Width - 1, FWorkBMP.Height - 1, Alpha1);
  end;

  if Effect = afxDoubleOutwardsVertical then begin
    FProgress := FProgress div 2;
    j := FWorkBMP.Height div 2;

    SetAlpha(0, 0, FWorkBMP.Width - 1, j - FProgress - 1 - 8, Alpha2);
    for i := 0 to 7 do
      SetAlpha(0, j - FProgress - 8 + i, FWorkBMP.Width - 1, j - FProgress - 8 + i, Alpha2 + Round(((255 - Alpha2) / 8) * (i + 1)));
    for i := 0 to 7 do
      if j - FProgress + i < j then
        SetAlpha(0, j - FProgress + i, FWorkBMP.Width - 1, j - FProgress + i, Alpha1 + Round(((255 - Alpha1) / 8) * (8 - (i + 1))));
    SetAlpha(0, j - FProgress + 8, FWorkBMP.Width - 1, j - 1, Alpha1);

    SetAlpha(0, j, FWorkBMP.Width - 1, j + FProgress - 8 - 1, Alpha1);
    for i := 0 to 7 do
      if j + FProgress - 8 + i >= j then
        SetAlpha(0, j + FProgress - 8 + i, FWorkBMP.Width - 1, j + FProgress - 8 + i, Alpha1 + Round(((255 - Alpha1) / 8) * (i + 1)));
    for i := 0 to 7 do
      SetAlpha(0, j + FProgress + i, FWorkBMP.Width - 1, j + FProgress + i, Alpha2 + Round(((255 - Alpha2) / 8) * (8 - (i + 1))));
    SetAlpha(0, j + FProgress + 8, FWorkBMP.Width - 1, FWorkBMP.Height - 1, Alpha2);
  end;

  if Effect = afxDoubleInwardsHorizontal then begin
    FProgress := FProgress div 2;
    j := FWorkBMP.Width div 2;

    SetAlpha(0, 0, FProgress - 1 - 8, FWorkBMP.Height - 1, Alpha1);
    for i := 0 to 7 do
      SetAlpha(FProgress - 8 + i, 0, FProgress - 8 + i, FWorkBMP.Height - 1, Alpha1 + Round(((255 - Alpha1) / 8) * (i + 1)));
    for i := 0 to 7 do
      if FProgress + i < j then
        SetAlpha(FProgress + i, 0, FProgress + i, FWorkBMP.Height - 1, Alpha2 + Round(((255 - Alpha2) / 8) * (8 - (i + 1))));
    SetAlpha(FProgress + 8, 0, j - 1, FWorkBMP.Height - 1, Alpha2);

    SetAlpha(j, 0, FWorkBMP.Width - FProgress - 1 - 8, FWorkBMP.Height - 1, Alpha2);
    for i := 0 to 7 do
      if FWorkBMP.Width - FProgress - 8 + i >= j then
        SetAlpha(FWorkBMP.Width - FProgress - 8 + i, 0, FWorkBMP.Width - FProgress - 8 + i, FWorkBMP.Height - 1, Alpha2 + Round(((255 - Alpha2) / 8) * (i + 1)));
    for i := 0 to 7 do
      SetAlpha(FWorkBMP.Width - FProgress + i, 0, FWorkBMP.Width - FProgress + i, FWorkBMP.Height - 1, Alpha1 + Round(((255 - Alpha1) / 8) * (8 - (i + 1))));
    SetAlpha(FWorkBMP.Width - FProgress + 8, 0, FWorkBMP.Width - 1, FWorkBMP.Height - 1, Alpha1);
  end;

  if Effect = afxDoubleOutwardsHorizontal then begin
    FProgress := FProgress div 2;
    j := FWorkBMP.Width div 2;

    SetAlpha(0, 0, j - FProgress - 1 - 8, FWorkBMP.Height - 1, Alpha2);
    for i := 0 to 7 do
      SetAlpha(j - FProgress - 8 + i, 0, j - FProgress - 8 + i, FWorkBMP.Height - 1, Alpha2 + Round(((255 - Alpha2) / 8) * (i + 1)));
    for i := 0 to 7 do
      if j - FProgress + i < j then
        SetAlpha(j - FProgress + i, 0, j - FProgress + i, FWorkBMP.Height - 1, Alpha1 + Round(((255 - Alpha1) / 8) * (8 - (i + 1))));
    SetAlpha(j - FProgress + 8, 0, j - 1, FWorkBMP.Height - 1, Alpha1);

    SetAlpha(j, 0, j + FProgress - 8 - 1, FWorkBMP.Height - 1, Alpha1);
    for i := 0 to 7 do
      if j + FProgress - 8 + i >= j then
        SetAlpha(j + FProgress - 8 + i, 0, j + FProgress - 8 + i, FWorkBMP.Height - 1, Alpha1 + Round(((255 - Alpha1) / 8) * (i + 1)));
    for i := 0 to 7 do
      SetAlpha(j + FProgress + i, 0, j + FProgress + i, FWorkBMP.Height - 1, Alpha2 + Round(((255 - Alpha2) / 8) * (8 - (i + 1))));
    SetAlpha(j + FProgress + 8, 0, FWorkBMP.Width - 1, FWorkBMP.Height - 1, Alpha2);
  end;
end;

initialization
  InitAlphaTables;
end.
